# ==== scan_M3_HC.py : enigma M3 en python

import sys
import enigmaM3
import string	
from bib_util import *
import HC_SCORE
import getopt
from WALHS import *
import random
import time
import signal, os

ALPHA=string.ascii_uppercase
# ===================================
def dechiffre(msg, e ):
    crypto = ""
    for car in msg:
        crypto += e.encode( car )
    return crypto
# ===================================
def affi_score( valeur, lg):
    if valeur > 1:
        valeur = valeur / lg
    ch = ("%6.4f" % valeur)
    return ch
# ===================================
def affiche( tb ):
    ch = ""
    for x in range( len(tb) ):
        ch += ALPHA[ tb[x] ]
    return ch
# ===================================
def extrait( debut, fin ):
    ch_extrait = ALPHA[ ALPHA.find(debut) : ALPHA.find(fin) + 1 ]
    return ch_extrait
# ===================================
def decode_grund( ch ):
    tb = ch.split(":")
    if len(tb) == 1:
        return (ch[0],ch[1],ch[2])
    gr_l = extrait( tb[0][0], tb[1][0])
    gr_m = extrait( tb[0][1], tb[1][1])
    gr_r = extrait( tb[0][2], tb[1][2])
    return (gr_l, gr_m, gr_r)
# ===================================
def decode_ring( ch ):
    tb = ch.split(":")
    if len(tb) == 1:
        return (ch[0], ch[1], ch[2])
    r_l = extrait( tb[0][0], tb[1][0])
    r_m = extrait( tb[0][1], tb[1][1])
    r_r = extrait( tb[0][2], tb[1][2])
    return (r_l, r_m, r_r)
# ===================================
def sig_handler(signum, frame):
    global last_time
    if time.time() - last_time < 10: sys.exit(0)
    print("[",walhs,RING,GRUND,int(time.time()-start_time)/60, end=' ')
    print(nb_dechif,"]")
    last_time = time.time()
# ===================================
def usage():
    print("usage:")
    print("  -h             Aide en ligne")
    print("  -c crypto      The cryptogram")
    print("  -v             Verbose (mode Debug)")
    print("  -W XYZ         The list of Walhzenagei. Examples:")
    print("     -W I,II,III:I,III,II")
    print("     MIN:  the 3 first rotors of Military  model [default]")
    print("     HEER: the 5 rotors of the Heer")
    print("     NAVY: the 8 rotors of the KriegsMarine")
    print("  -G XXX:ZZZ     The Grund domain. Examples: ")
    print("     -G AAA:ZZZ [by default], -G DGJ, AGA:AGZ, ...")
    print("  -R XXX:ZZZZ     The Ringstellung domain (AAA by default)")
    print("  -M method      The method (IC,UNI,BIG,TRI,QUA,PLN,CRIB), ")
    print("                 default: IC")
    print("  -m file_method The file used with method (german by default)")
    print("  -U UKW         The Reflector, by default B")
    print("  -s steckers    The Steckers, ex: AH:JQ:LP:EZ:RT:NF, nothing by default")
    print("  -S seuil       Print solutions above threshold (by default print max values)")
    print("  -P             Print maximum values (by default)")
    print("  -p             Print values above threshold ")
    print("                 [not compatible with -p])")
    print("Ctrl-C prints the progression, two Ctrl-C stops the program")
# ------------------ DEBUT PROGRAMME ---------------------
# ----- on traite les options
try:
    opts, args = getopt.getopt(sys.argv[1:],
        "hvc:W:G:R:s:M:m:U:S:P", ["help", "verbose", "crypto=", 
        "walzenlage=", "grund=", "ring=", "steckers=", "method=",
        "file_method=", "ukw=", "threshold=", "noprintmax=" ])
except getopt.GetoptError as err:
    usage()
    sys.exit(2)

DEBUG=False
PRINT_MAX=True
msg = ""
GRUND_UKW    = string.ascii_uppercase
GRUND_LEFT   = string.ascii_uppercase
GRUND_MIDDLE = string.ascii_uppercase
GRUND_RIGHT  = string.ascii_uppercase
RING_UKW  = string.ascii_uppercase
RING_LEFT = string.ascii_uppercase
RING_MIDDLE = string.ascii_uppercase
RING_RIGHT  = string.ascii_uppercase
WALZENLAGE = "MIN"
METHOD = "IC"
FILE_METHOD = "german"
UKW = 'B' 
STECKERS = ""
THRESHOLD = 0

for o, a in opts:
    if o == "-v":
        DEBUG = True
    elif o in ("-h", "--help"):
        usage()
        sys.exit()
    elif o in ("-c", "--crypto"):
        msg = litfic( a )
    elif o in ("-W", "--walzenlage"):
        WALZENLAGE = a
    elif o in ("-G", "--grund"):
        (GRUND_LEFT, GRUND_MIDDLE, GRUND_RIGHT) = decode_grund(a)
    elif o in ("-R", "--ring"):
        (RING_LEFT, RING_MIDDLE, RING_RIGHT) = decode_ring(a)
    elif o in ("-M", "--method"):
        METHOD = a 
    elif o in ("-m", "--file_method"):
        FILE_METHOD = a
    elif o in ("-U", "--ukw"):
        UKW = a
    elif o in ("-s", "--steckers"):
        STECKERS = a
    elif o in ("-S", "--treshold"):
        THRESHOLD = float(a)
    elif o in ("-P", "--noprintmax"):
        PRINT_MAX = False
    else:
        assert False, "unhandled option"
if msg == "":
    usage()
    sys.exit(1)
# ---------------------------
start_time = time.time() ; last_time = 0
signal.signal(signal.SIGINT, sig_handler)
HC_SCORE.score_init[ METHOD ]( FILE_METHOD )
MAX_SCORE = 0
nb_dechif = 0

for walhs in LES_WALHS( WALZENLAGE ):
 tb = walhs.split(",")
 LEFT = tb[0]; MIDDLE = tb[1]; RIGHT = tb[2] 
 RING='AAA'; GRUND='AAA'
 c = enigmaM3.Enigma(UKW,LEFT,MIDDLE,RIGHT,STECKERS,RING,GRUND,DEBUG)
 for g_l in GRUND_LEFT:
  for g_m in GRUND_MIDDLE:
   for g_r in GRUND_RIGHT:
    for r_l in RING_LEFT:
     for r_m in RING_MIDDLE:
      for r_r in RING_RIGHT:
         GRUND = g_l + g_m + g_r
         RING =  r_l + r_m + r_r
         c.setRing( RING )
         c.setGrund( GRUND )
         plain = dechiffre( msg, c )
         CUR_HIGH_SCORE = score = HC_SCORE.score[ METHOD ](plain)
         nb_dechif += 1
       
         now = time.time() 
         elaps_mn = int( now - start_time ) / 60
         elaps_sec= int( now - start_time ) % 60
         if PRINT_MAX and CUR_HIGH_SCORE > MAX_SCORE:
           MAX_WALHZ = walhs
           MAX_SCORE = CUR_HIGH_SCORE
           MAX_PLAIN = plain
           MAX_KEY   = UKW + " " + " " + RING + " " + GRUND  
           if CUR_HIGH_SCORE >= THRESHOLD:
             print(">>",MAX_WALHZ, MAX_KEY,":", \
              affi_score( MAX_SCORE, len(msg) ),\
              "-", MAX_PLAIN[0:50], nb_dechif, \
              int(elaps_mn), "mn", elaps_sec, "sec")
           continue
           #print(">>>", PRINT_MAX, CUR_HIGH_SCORE, TRESHOLD)
         if not PRINT_MAX and CUR_HIGH_SCORE > THRESHOLD: 
            key = UKW + " " + " " + RING + " " + GRUND  
            print(">", walhs, key, ":", \
              affi_score( CUR_HIGH_SCORE, len(msg) ),\
              "-", plain[0:50], nb_dechif, \
              int(elaps_mn), "mn", elaps_sec, "sec")
            continue

